import asyncio
import logging
from typing import Optional, List, Union, Pattern, Dict, Any

from khl import Message, Client
from .command import Command
from .exception import TypeEHandler
from .lexer import Lexer, DefaultLexer
from .parser import Parser
from .rule import TypeRule

log = logging.getLogger(__name__)


class CommandManager:
    """aggregates commands as a system, and provide API as a whole"""
    _cmd_map: Dict[str, Command]

    def __init__(self):
        self._cmd_map = {}

    # why disable duplicate-code:
    # the duplicated code is the param list, used to provide auto complete/coding hints in IDE
    # pylint: disable = duplicate-code
    def __call__(
            self,
            name: str = '',
            *,
            help: str = '',
            desc: str = '',
            aliases: List[str] = (),
            prefixes: List[str] = ('/', ),
            regex: Union[str, Pattern] = '',
            lexer: Lexer = None,
            parser: Parser = None,
            rules: List[TypeRule] = (),
            exc_handlers: Dict[Any, TypeEHandler] = None,
            case_sensitive: bool = True
    ):
        """
        decorator, wrap a function in Command and register it on current Bot

        :param name: the name of this Command, also used to trigger command in DefaultLexer
        :param aliases: (DefaultLexer only) you can also trigger the command with aliases
        :param prefixes: (DefaultLexer only) command prefix, default use '/'
        :param regex: (RELexer only) pattern for the command
        :param help: detailed manual
        :param desc: short introduction
        :param lexer: (Advanced) explicitly set the lexer
        :param parser: (Advanced) explicitly set the parser
        :param rules: only be executed if all rules are met
        :return: wrapped Command
        """
        args = {
            'help': help,
            'desc': desc,
            'aliases': aliases,
            'prefixes': prefixes,
            'regex': regex,
            'lexer': lexer,
            'parser': parser,
            'rules': rules,
            'exc_handlers': exc_handlers,
            'case_sensitive': case_sensitive
        }

        return lambda func: self.add(Command.command(name, **args)(func))

    def add(self, command: Command) -> Command:
        """register the cmd on current Bot

        :param command: the Command going to be registered
        :return: the cmd
        """
        self[command.name] = command
        return command

    def get(self, name: str) -> Optional[Command]:
        """get command by name"""
        return self[name]

    def pop(self, name: str) -> Optional[Command]:
        """pop a command from Manager, return None if not exist"""
        cmd = self[name]
        if cmd:
            del self._cmd_map[name]
        return cmd

    async def handle(self, loop, client: Client, msg: Message, filter_args: dict):
        """pass msg into all commands in self, handle it concurrently"""
        for cmd in self._cmd_map.values():
            asyncio.ensure_future(cmd.handle(msg, client, filter_args), loop=loop)

    def update_prefixes(self, *prefixes: str) -> List[Command]:
        """update command prefixes in the Manager if command uses DefaultLexer

        :return: updated commands
        """
        prefixes = set(prefixes)
        updated = []
        for _, cmd in self.items():
            if not isinstance(cmd.lexer, DefaultLexer):
                continue
            cmd.lexer.prefixes = prefixes
            updated.append(cmd)
        return updated

    def __setitem__(self, name: str, cmd: Command):
        if cmd.name in self._cmd_map:
            raise ValueError(f'cmd: {cmd.name} already exists')
        self._cmd_map[cmd.name] = cmd
        log.debug(f'command: {cmd.name} added')

    def __getitem__(self, item) -> Optional[Command]:
        return self._cmd_map.get(item, None)

    def __iter__(self):
        return iter(self._cmd_map)

    def items(self):
        """all commands in self, keyed by command name"""
        return self._cmd_map.items()



"""command

command is an abstraction of code-user reaction mode:
user send a message, the code check if it matches command scheme, and handle it if matches"""
from .lexer import Lexer, DefaultLexer, RELexer
from .parser import Parser

from .exception import Exceptions
from .rule import Rule

from .command import Command

from .manager import CommandManager



"""abstraction of khl concept channel: where messages flow in"""
import json
from abc import ABC, abstractmethod
from typing import Union, List, Dict, Optional

from . import api
from .gateway import Requestable, Gateway
from .interface import LazyLoadable
from .permission import ChannelPermission
from .role import Role
from ._types import MessageTypes, ChannelTypes, SlowModeTypes
from .user import User
from .util import unpack_value


class Channel(LazyLoadable, Requestable, ABC):
    """
    Interface, represents a channel where messages flowing
    """
    type: ChannelTypes

    @abstractmethod
    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        """
        send a msg to the channel
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def id(self) -> str:
        """the channel's id

        this field should be protected, thus only exported a read-only prop"""
        raise NotImplementedError


class PublicChannel(Channel, ABC):
    """the channels in guild, in contrast to PrivateChannel(private chat)"""
    name: str
    user_id: str
    guild_id: str
    topic: str
    is_category: int
    parent_id: str
    level: int
    permission: ChannelPermission

    def __init__(self, **kwargs):
        self._id: str = kwargs.get('id')
        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')
        self._update_fields(**kwargs)

    @property
    def id(self) -> str:
        return self._id

    def _update_fields(self, **kwargs):
        self.name: str = kwargs.get('name')
        self.user_id: str = kwargs.get('user_id')
        self.guild_id: str = kwargs.get('guild_id')
        self.topic: str = kwargs.get('topic')
        self.is_category: int = kwargs.get('is_category')
        self.parent_id: str = kwargs.get('parent_id')
        self.level: int = kwargs.get('level')
        self.type: ChannelTypes = kwargs.get('type') and ChannelTypes(kwargs.get('type'))
        self.permission: ChannelPermission = ChannelPermission(**kwargs)

    async def load(self):
        self._update_fields(**(await self.gate.exec_req(api.Channel.view(self.id))))
        self._loaded = True

    async def update(self, name: str = None, topic: str = None, slow_mode: Union[int, SlowModeTypes] = None) -> dict:
        """
        update channel's settings
        """
        params = {'channel_id': self.id}
        if name is not None:
            params['name'] = name
        if topic is not None:
            params['topic'] = topic
        if slow_mode is not None:
            params['slow_mode'] = unpack_value(slow_mode)
        rt = await self.gate.exec_req(api.Channel.update(**params))
        await self.load()
        return rt

    async def fetch_permission(self, force_update: bool = True) -> ChannelPermission:
        """fetch permission setting of the channel"""
        if force_update or not self.permission.loaded:
            await self.permission.load()
        return self.permission

    async def create_user_permission(self, target: Union[User, str]):
        """create a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'user_id'
        v = target.id if isinstance(target, User) else target
        d = await self.gate.exec_req(api.ChannelRole.create(channel_id=self.id, type=t, value=v))
        self.permission.loaded = False
        return d

    async def update_user_permission(self, target: Union[User, str], allow: int = 0, deny: int = 0) -> Role:
        """update a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'user_id'
        v = target.id if isinstance(target, User) else target
        return await self.gate.exec_req(
            api.ChannelRole.update(channel_id=self.id, type=t, value=v, allow=allow, deny=deny))

    async def delete_user_permission(self, target: Union[User, str]):
        """delete a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'user_id'
        v = target.id if isinstance(target, User) else target
        return await self.gate.exec_req(api.ChannelRole.delete(channel_id=self.id, type=t, value=v))

    async def create_role_permission(self, target: Union[Role, str]):
        """create a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'role_id'
        v = target.id if isinstance(target, Role) else target
        d = await self.gate.exec_req(api.ChannelRole.create(channel_id=self.id, type=t, value=v))
        self.permission.loaded = False
        return d

    async def update_role_permission(self, target: Union[Role, str], allow: int = 0, deny: int = 0) -> Role:
        """update a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'role_id'
        v = target.id if isinstance(target, Role) else target
        return await self.gate.exec_req(
            api.ChannelRole.update(channel_id=self.id, type=t, value=v, allow=allow, deny=deny))

    async def delete_role_permission(self, target: Union[Role, str]):
        """delete a customized permission setting entry

        for permission setting entry, please refer to `ChannelPermission`"""
        t = 'role_id'
        v = target.id if isinstance(target, Role) else target
        return await self.gate.exec_req(api.ChannelRole.delete(channel_id=self.id, type=t, value=v))


class PublicTextChannel(PublicChannel):
    """
    `Standard Object`

    Text chat channels in guild
    """
    slow_mode: int

    def _update_fields(self, **kwargs):
        super()._update_fields(**kwargs)
        self.slow_mode: int = kwargs.get('slow_mode')

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, temp_target_id: str = '', **kwargs):
        """
        send a msg to the channel

        ``temp_target_id`` is available in PublicTextChannel, so ``send()`` is overloaded here
        """
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            type = MessageTypes.CARD
            content = json.dumps(content)
        type = type if type is not None else MessageTypes.KMD

        # merge params
        kwargs['target_id'] = self.id
        kwargs['content'] = content
        kwargs['type'] = type.value
        if temp_target_id:
            kwargs['temp_target_id'] = temp_target_id

        return await self.gate.exec_req(api.Message.create(**kwargs))


class PublicVoiceChannel(PublicChannel):
    """
    Voice chat channel in guild

    a placeholder now for future design/adaption
    """

    async def send(self, content: Union[str, List], **kwargs):
        raise TypeError('now there is no PublicVoiceChannel, *hey dude we have a pkg from future*')

    async def move_user(self, *users: Union[User, str]):
        """move users to this public voice channel"""
        user_ids = [u.id if isinstance(u, User) else u for u in users]
        return await self.gate.exec_req(api.Channel.moveUser(target_id=self.id, user_ids=user_ids))


def public_channel_factory(_gate_: Gateway, **kwargs) -> Optional[PublicChannel]:
    """factory function to build a channel object"""
    kwargs['type'] = kwargs['type'] if isinstance(kwargs['type'], ChannelTypes) else ChannelTypes(kwargs['type'])
    if kwargs['type'] == ChannelTypes.TEXT:
        return PublicTextChannel(**kwargs, _gate_=_gate_)
    if kwargs['type'] == ChannelTypes.VOICE:
        return PublicVoiceChannel(**kwargs, _gate_=_gate_)
    raise ValueError(f'unsupported channel type: {kwargs["type"]}: {kwargs}')


class PrivateChannel(Channel):
    """Private chat channel

    a private channel associates the code and another user(called as target in the following)"""

    code: str
    last_read_time: int
    latest_msg_time: int
    unread_count: int
    is_friend: bool
    is_blocked: bool
    is_target_blocked: bool
    target_info: Dict

    def __init__(self, **kwargs):
        self.code: str = kwargs.get('code')
        self.last_read_time: int = kwargs.get('last_read_time')
        self.latest_msg_time: int = kwargs.get('latest_msg_time')
        self.unread_count: int = kwargs.get('unread_count')
        self.is_friend: bool = kwargs.get('is_friend')
        self.is_blocked: bool = kwargs.get('is_blocked')
        self.is_target_blocked: bool = kwargs.get('is_target_blocked')
        self.target_info: Dict = kwargs.get('target_info')

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_')

    async def load(self):
        pass

    @property
    def id(self) -> str:
        return self.target_user_id

    @property
    def target_user_id(self) -> str:
        """prop, the target's id"""
        return self.target_info.get('id') if self.target_info else None

    @property
    def target_user_name(self) -> str:
        """prop, the target's name"""
        return self.target_info.get('username') if self.target_info else None

    @property
    def is_target_user_online(self) -> bool:
        """prop, is the target online"""
        return self.target_info.get('online') if self.target_info else None

    @property
    def target_user_avatar(self) -> str:
        """prop, the target's avatar"""
        return self.target_info.get('avatar') if self.target_info else None

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        return await User(id=self.id, _gate_=self.gate).send(content, type=type, **kwargs)



"""components in Element category"""
from typing import Union, Dict

from .interface import Types, _Common


class Element:
    """components in Element category

    the most basic content delivers, usually contained in module"""

    class Text(_Common):
        """Text Element"""
        content: str
        emoji: bool

        def __init__(self, content: str, type: Union[Types.Text, str] = Types.Text.PLAIN, emoji: bool = True):
            if isinstance(type, str):
                type = Types.Text(type)  # check if type in Type.Text
            self._type = type.value
            self.content = content
            self.emoji = emoji
            super().__init__(Types.Theme.NA, Types.Size.NA)

        @property
        def _repr(self) -> Union[Dict, str]:
            # 「为了方便书写，所有plain-text的使用处可以简单的用字符串代替。」
            if self._type == Types.Text.PLAIN.value and self.emoji:
                return self.content
            result = self._gen_dict(['type', 'content'])
            if self._type == Types.Text.PLAIN.value:
                result['emoji'] = self.emoji
            return result

    class Image(_Common):
        """Image Element"""
        _type = 'image'
        src: str
        alt: str
        circle: bool

        def __init__(self, src: str, alt: str = '', circle: bool = False, size: Union[Types.Size, str] = Types.Size.LG):
            self.src = src
            self.alt = alt
            self.circle = circle
            if isinstance(size, str):
                size = Types.Size(size)
            super().__init__(Types.Theme.NA, size)

        @property
        def _repr(self) -> Dict:
            return self._gen_dict(['type', 'src', 'alt', 'size', 'circle'])

    class Button(_Common):
        """Button Element

        can react to user's click"""
        _type = 'button'
        _click: Types.Click
        text: 'Element.Text'
        value: str

        def __init__(self,
                     text: Union['Element.Text', str],
                     value: str = '',
                     click: Union[Types.Click, str] = Types.Click.RETURN_VAL,
                     theme: Union[Types.Theme, str, None] = None):
            self.text = text
            self._click = click if isinstance(click, Types.Click) else Types.Click(click)
            self.value = value
            super().__init__(theme, Types.Size.NA)

        @property
        def click(self) -> str:
            """there is two types of actions for button to react when user clicks, which depends on the `Button.click`:

            1. `Button.click` == Types.Click.RETURN_VAL:
            returns the `Button.value` back to bot

            2. `Button.click` == Types.Click.LINK:
            khl client will open a browser tab for user, with link = `Button.value`"""
            return self._click.value

        @click.setter
        def click(self, value: Union[Types.Click, str]):
            self._click = value if isinstance(value, Types.Click) else Types.Click(value)

        @property
        def _repr(self) -> Union[Dict, str]:
            return self._gen_dict(['type', 'value', 'click', 'text', 'theme'])



"""card message: wrapper for easier card message sending"""
from .card import Card
from .interface import _get_repr


class CardMessage(list):
    """
    can be cast to a list, the most outside wrapper for card message
    """

    def __init__(self, *cards: Card):
        super().__init__()
        self.extend(cards)

    def __iter__(self):
        """hack for JSON serialization"""
        return iter([_get_repr(i) for i in self[:]])



from abc import ABC, abstractmethod
from typing import Any, List, Dict, Union

import json

from . import api
from .channel import PublicTextChannel, PrivateChannel
from .context import Context
from .gateway import Requestable
from .guild import Guild
from ._types import MessageTypes, ChannelPrivacyTypes, EventTypes
from .user import User

class RawMessage(ABC):
    """
    Basic and common features of kinds of messages.

    now support Message kinds:
        1. Message (sent by users, those normal chats such as TEXT/IMG etc.)
        2. Event (sent by system, such as notifications and broadcasts)
    """

    _type: int
    _channel_type: str
    target_id: str
    author_id: str
    content: str
    _msg_id: str
    msg_timestamp: int
    nonce: str
    extra: Any

    def __init__(self, **kwargs):
        self._msg_id = kwargs.get('msg_id')
        self._type = kwargs.get('type')
        self._channel_type = kwargs.get('channel_type')
        self.target_id = kwargs.get('target_id')
        self.author_id = kwargs.get('author_id')
        self.content = kwargs.get('content')
        self.msg_timestamp = kwargs.get('msg_timestamp')
        self.nonce = kwargs.get('nonce')
        self.extra = kwargs.get('extra', {})

    @property
    def id(self) -> str:
        """message's id"""
        return self._msg_id

    @property
    def type(self) -> MessageTypes:
        """message's type, refer to MessageTypes for enum detail"""
        return MessageTypes(self._type)

    @property
    def channel_type(self) -> ChannelPrivacyTypes:
        """type of the channel where the message in"""
        return ChannelPrivacyTypes(self._channel_type)


class Message(RawMessage, Requestable, ABC):
    """
    Represents the messages sent by user.

    Because it has source and context, we can interact with its sender and context via it.

    now there are two types of message:
        1. ChannelMessage: sent in a guild channel
        2. PrivateMessage: sent in a private chat
    """
    _ctx: Context

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.gate = kwargs.get('_gate_', None)
        self._author = User(**self.extra['author'], _gate_=self.gate, _lazy_loaded_=True)

    @property
    def author(self) -> User:
        """message author"""
        return self._author

    @property
    def ctx(self) -> Context:
        """message context: channel, guild etc."""
        return self._ctx

    @abstractmethod
    async def add_reaction(self, emoji: str):
        """add emoji to msg's reaction list

        https://developer.kaiheila.cn/doc/http/message#%E7%BB%99%E6%9F%90%E4%B8%AA%E6%B6%88%E6%81%AF%E6%B7%BB%E5%8A%A0%E5%9B%9E%E5%BA%94

        https://developer.kaiheila.cn/doc/http/direct-message#%E7%BB%99%E6%9F%90%E4%B8%AA%E6%B6%88%E6%81%AF%E6%B7%BB%E5%8A%A0%E5%9B%9E%E5%BA%94

        :param emoji: 😘
        """

    @abstractmethod
    async def delete_reaction(self, emoji: str, user: User):
        """delete emoji from msg's reaction list

        https://developer.kaiheila.cn/doc/http/message#%E5%88%A0%E9%99%A4%E6%B6%88%E6%81%AF%E7%9A%84%E6%9F%90%E4%B8%AA%E5%9B%9E%E5%BA%94

        https://developer.kaiheila.cn/doc/http/direct-message#%E5%88%A0%E9%99%A4%E6%B6%88%E6%81%AF%E7%9A%84%E6%9F%90%E4%B8%AA%E5%9B%9E%E5%BA%94

        :param emoji: 😘
        :param user: whose reaction, delete others added reaction requires channel msg admin permission
        """

    @abstractmethod
    async def update(self, content: Union[str, List], quote: str = None, temp_target_id: str = None):
        """Update the message content, the type of content should be same the original type of the message
        https://developer.kookapp.cn/doc/http/message#%E6%9B%B4%E6%96%B0%E9%A2%91%E9%81%93%E8%81%8A%E5%A4%A9%E6%B6%88%E6%81%AF
        https://developer.kookapp.cn/doc/http/direct-message#%E6%9B%B4%E6%96%B0%E7%A7%81%E4%BF%A1%E8%81%8A%E5%A4%A9%E6%B6%88%E6%81%AF
        :param content: updated content, its type should be same as the original type
        :param quote: the id of the message that will be quoted
        :param temp_target_id: only update the message for which user
        """

    async def reply(self,
                    content: Union[str, List] = '',
                    use_quote: bool = True,
                    *,
                    type: MessageTypes = None,
                    **kwargs):
        """
        reply to a msg, content can also be a card
        """
        if use_quote:
            kwargs['quote'] = self.id

        return await self.ctx.channel.send(content, type=type, **kwargs)

    async def delete(self):
        """delete the message, permission required"""
        return await self.gate.exec_req(api.Message.delete(msg_id=self.id))


class PublicMessage(Message):
    """
    Messages sent in a `PublicTextChannel`
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        channel = PublicTextChannel(id=self.target_id, name=self.extra['channel_name'], _gate_=self.gate)
        guild = Guild(id=self.extra['guild_id'], _gate_=self.gate)
        self._ctx = Context(channel=channel, guild=guild, _gate_=self.gate)

    @property
    def guild(self) -> Guild:
        """the guild where the message in"""
        return self.ctx.guild

    @property
    def channel(self) -> PublicTextChannel:
        """the channel where the message in"""
        if isinstance(self.ctx.channel, PublicTextChannel):
            return self.ctx.channel
        raise ValueError('PublicMessage should be placed in PublicTextChannel')

    @property
    def mention(self) -> List[str]:
        """the message mentioned(also call as at/tagged) users' id"""
        return self.extra['mention']

    @property
    def mention_all(self) -> bool:
        """if the message mentioned(also call as at/tagged) all"""
        return self.extra['mention_all']

    @property
    def mention_roles(self) -> List:
        """the message mentioned(also call as at/tagged) roles' id"""
        return self.extra['mention_roles']

    @property
    def mention_here(self) -> bool:
        """if the message mentioned(also call as at/tagged) all online users in the channel"""
        return self.extra['mention_here']

    async def add_reaction(self, emoji: str):
        return await self.gate.exec_req(api.Message.addReaction(msg_id=self.id, emoji=emoji))

    async def delete_reaction(self, emoji: str, user: User = None):
        req = api.Message.deleteReaction(msg_id=self.id, emoji=emoji, user_id=user.id if user else '')
        return await self.gate.exec_req(req)

    async def update(self, content: Union[str, List], quote: str = None, temp_target_id: str = None):
        if isinstance(content, List):
            content = json.dumps(content)
        params = {'msg_id': self.id, 'content': content}
        if quote is not None:
            params['quote'] = quote
        if temp_target_id is not None:
            params['temp_target_id'] = temp_target_id
        return await self.gate.exec_req(api.Message.update(**params))

    async def reply(self,
                    content: Union[str, List] = '',
                    use_quote: bool = True,
                    is_temp: bool = False,
                    *,
                    type: MessageTypes = None,
                    **kwargs):
        return await super().reply(content,
                                   use_quote,
                                   temp_target_id=self.author_id if is_temp else '',
                                   type=type,
                                   **kwargs)


class PrivateMessage(Message):
    """
    Messages sent in a `PrivateChannel`
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._channel = PrivateChannel(code=self.extra['code'], target_info=self.extra['author'], _gate_=self.gate)
        self._ctx = Context(channel=self._channel, _gate_=self.gate)

    @property
    def chat_code(self) -> str:
        """the chat code of this private chat"""
        return self.extra['code']

    @property
    def channel(self) -> PrivateChannel:
        """the message's channel"""
        return self._channel

    async def add_reaction(self, emoji: str):
        return await self.gate.exec_req(api.DirectMessage.addReaction(msg_id=self.id, emoji=emoji))

    async def delete_reaction(self, emoji: str, user: User = None):
        req = api.DirectMessage.deleteReaction(msg_id=self.id, emoji=emoji, user_id=user.id if user else '')
        return await self.gate.exec_req(req)

    async def update(self, content: Union[str, List], quote: str = None, _: str = None):
        if isinstance(content, List):
            content = json.dumps(content)
        params = {'msg_id': self.id, 'content': content}
        if quote is not None:
            params['quote'] = quote
        return await self.gate.exec_req(api.DirectMessage.update(**params))


class Event(RawMessage):
    """sent by system, opposites to Message, carries various types of payload"""

    @property
    def event_type(self) -> EventTypes:
        """type of the event, refer to EventTypes for enum detail"""
        return EventTypes(self.extra['type'])

    @property
    def body(self) -> Dict:
        """event body, a dict, refer to official docs with the event_type for the actual struct

        docs: https://developer.kaiheila.cn/doc/event/event-introduction"""
        return self.extra['body']



"""implementation of bot"""
import asyncio
import logging
import warnings
from pathlib import Path
from typing import Dict, Callable, List, Optional, Union, Coroutine, IO

from .. import AsyncRunnable  # interfaces
from .. import Cert, HTTPRequester, WebhookReceiver, WebsocketReceiver, Gateway, Client  # net related
from .. import MessageTypes, EventTypes, SlowModeTypes, SoftwareTypes  # types
from .. import User, Channel, PublicChannel, Guild, Event, Message  # concepts
from ..command import CommandManager
from ..game import Game
from ..task import TaskManager

log = logging.getLogger(__name__)

TypeEventHandler = Callable[['Bot', Event], Coroutine]
TypeMessageHandler = Callable[[Message], Coroutine]
TypeStartupHandler = Callable[['Bot'], Coroutine]
TypeShutdownHandler = Callable[['Bot'], Coroutine]


class Bot(AsyncRunnable):
    """
    Represents a entity that handles msg/events and interact with users/khl server in manners that programmed.
    """
    # components
    client: Client
    command: CommandManager
    task: TaskManager

    # flags
    _is_running: bool

    # internal containers
    _me: Optional[User]
    _event_index: Dict[EventTypes, List[TypeEventHandler]]

    _startup_index: List[TypeStartupHandler]
    _shutdown_index: List[TypeShutdownHandler]

    def __init__(self,
                 token: str = '',
                 *,
                 cert: Cert = None,
                 client: Client = None,
                 gate: Gateway = None,
                 out: HTTPRequester = None,
                 compress: bool = True,
                 port=5000,
                 route='/khl-wh'):
        """
        The most common usage: ``Bot(token='xxxxxx')``

        That's enough.

        :param cert: used to build requester and receiver
        :param client: the bot relies on
        :param gate: the client relies on
        :param out: the gate's component
        :param compress: used to tune the receiver
        :param port: used to tune the WebhookReceiver
        :param route: used to tune the WebhookReceiver
        """
        if not token and not cert:
            raise ValueError('require token or cert')

        self._init_client(cert or Cert(token=token), client, gate, out, compress, port, route)
        self._register_client_handler()

        self.command = CommandManager()

        self.task = TaskManager()

        self._is_running = False
        self._event_index = {}
        self._startup_index = []
        self._shutdown_index = []

    def _init_client(self, cert: Cert, client: Client, gate: Gateway, out: HTTPRequester, compress: bool, port, route):
        """
        construct self.client from args.

        you can init client with kinds of filling ways,
        so there is a priority in the rule: client > gate > out = compress = port = route

        :param cert: used to build requester and receiver
        :param client: the bot relies on
        :param gate: the client relies on
        :param out: the gate's component
        :param compress: used to tune the receiver
        :param port: used to tune the WebhookReceiver
        :param route: used to tune the WebhookReceiver
        :return:
        """
        if client:
            self.client = client
            return
        if gate:
            self.client = Client(gate)
            return

        # client and gate not in args, build them
        _out = out if out else HTTPRequester(cert)
        if cert.type == Cert.Types.WEBSOCKET:
            _in = WebsocketReceiver(cert, compress)
        elif cert.type == Cert.Types.WEBHOOK:
            _in = WebhookReceiver(cert, port=port, route=route, compress=compress)
        else:
            raise ValueError(f'cert type: {cert.type} not supported')

        self.client = Client(Gateway(_out, _in))

    def _register_client_handler(self):
        # text and kmd -> msg
        msg_handler = self._make_msg_handler()
        self.client.register(MessageTypes.TEXT, msg_handler)
        self.client.register(MessageTypes.KMD, msg_handler)

        # sys -> event
        self.client.register(MessageTypes.SYS, self._make_event_handler())

    def _make_msg_handler(self) -> Callable:
        """
        construct a function to receive msg from client, and interpret it with _cmd_index
        """

        async def handler(msg: Message):
            await self.command.handle(self.loop, self.client, msg, {Message: msg, Bot: self})

        return handler

    def _make_event_handler(self) -> Callable:

        async def handler(event: Event):
            if event.event_type not in self._event_index:
                return
            if not self._event_index[event.event_type]:
                return
            for event_handler in self._event_index[event.event_type]:
                await event_handler(self, event)

        return handler

    def add_event_handler(self, type: EventTypes, handler: TypeEventHandler):
        """add an event handler function for EventTypes `type`"""
        if type not in self._event_index:
            self._event_index[type] = []
        self._event_index[type].append(handler)
        log.debug(f'event_handler {handler.__qualname__} for {type} added')
        return handler

    def add_message_handler(self, handler: TypeMessageHandler, *except_type: MessageTypes):
        """`except_type` is an exclusion list"""
        for type in MessageTypes:
            if type not in except_type:
                self.client.register(type, handler)

    def on_event(self, type: EventTypes):
        """decorator, register a function to handle events of the type"""

        def dec(func: TypeEventHandler):
            self.add_event_handler(type, func)

        return dec

    def on_message(self, *except_type: MessageTypes):
        """
        decorator, register a function to handle messages
        :param except_type: excepted types
        """

        def dec(func: TypeMessageHandler):
            self.add_message_handler(func, *set(except_type + (MessageTypes.SYS, )))

        return dec

    def on_startup(self, func: TypeStartupHandler):
        """decorator, register a function to handle bot start"""

        self._startup_index.append(func)

        return func

    def on_shutdown(self, func: TypeShutdownHandler):
        """decorator, register a function to handle bot stop"""

        self._shutdown_index.append(func)

        return func

    async def fetch_me(self, force_update: bool = False) -> User:
        """fetch detail of the bot it self as a ``User``

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_me()`"""
        warnings.warn("deprecated, alternative: bot.client.fetch_me()", DeprecationWarning, stacklevel=2)
        return await self.client.fetch_me(force_update)

    @property
    def me(self) -> User:
        """
        get bot itself data

        CAUTION: please call ``await fetch_me()`` first to load data from khl server

        designed as 'empty-then-fetch' will break the rule 'net-related is async'

        :returns: the bot's underlying User

        .. deprecated-removed:: 0.3.0 0.4.0
            use await :func:`.client.fetch_me()`
        """
        warnings.warn("deprecated, alternative: bot.client.fetch_me(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return self.client.me

    async def fetch_public_channel(self, channel_id: str) -> PublicChannel:
        """channel id -> :class:`PublicChannel` object(public channel only),
        fetch details of a public channel from khl

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_public_channel()`"""
        warnings.warn("deprecated, alternative: bot.client.fetch_public_channel(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.fetch_public_channel(channel_id)

    async def fetch_user(self, user_id: str) -> User:
        """user id -> :class:`User` object, fetch user info from khl

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_user()`"""
        warnings.warn("deprecated, alternative: bot.client.fetch_user(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.fetch_user(user_id)

    async def delete_channel(self, channel: Union[Channel, str]):
        """delete a channel, permission required

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.delete_channel()`"""
        warnings.warn("deprecated, alternative: bot.client.delete_channel(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.delete_channel(channel)

    async def fetch_guild(self, guild_id: str) -> Guild:
        """guild id -> :class:`Guild` object, fetch details of a guild from khl

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_guild()`"""
        warnings.warn("deprecated, alternative: bot.client.fetch_guild(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.fetch_guild(guild_id)

    async def list_guild(self) -> List[Guild]:
        """list guilds the bot joined

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_guild_list()`
        """
        warnings.warn("deprecated, alternative: bot.client.fetch_guild_list(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.fetch_guild_list()

    async def send(self,
                   target: Channel,
                   content: Union[str, List],
                   *,
                   type: MessageTypes = None,
                   temp_target_id: str = '',
                   **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.send()`"""
        warnings.warn("deprecated, alternative: bot.client.send(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.send(target, content, type=type, temp_target_id=temp_target_id, **kwargs)

    async def upload_asset(self, file: Union[IO, str, Path]) -> str:
        """upload ``file`` to khl, and return the url to the file, alias for ``create_asset``

        if ``file`` is a str or Path, ``open(file, 'rb')`` will be called to convert it into IO

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.create_asset()`"""
        warnings.warn("deprecated, alternative: bot.client.create_asset(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.create_asset(file)

    async def create_asset(self, file: Union[IO, str, Path]) -> str:
        """upload ``file`` to khl, and return the url to the file

        if ``file`` is a str or Path, ``open(file, 'rb')`` will be called to convert it into IO

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.create_asset()`"""
        warnings.warn("deprecated, alternative: bot.client.create_asset(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.create_asset(file)

    async def kickout(self, guild: Union[Guild, str], user: Union[User, str]):
        """kick ``user`` out from ``guild``

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.kickout()`"""
        warnings.warn("deprecated, alternative: bot.client.kickout(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.kickout(guild, user)

    async def leave(self, guild: Union[Guild, str]):
        """leave from ``guild``

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.leave()`"""
        warnings.warn("deprecated, alternative: bot.client.leave(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.leave(guild)

    async def add_reaction(self, msg: Message, emoji: str):
        """add emoji to msg's reaction list

        wraps `Message.add_reaction`

        :param msg: accepts `Message`
        :param emoji: 😘

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.add_reaction()`
        """
        warnings.warn("deprecated, alternative: bot.client.add_reaction(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.add_reaction(msg, emoji)

    async def delete_reaction(self, msg: Message, emoji: str, user: User = None):
        """delete emoji from msg's reaction list

        wraps `Message.delete_reaction`

        :param msg: accepts `Message`
        :param emoji: 😘
        :param user: whose reaction, delete others added reaction requires channel msg admin permission

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.delete_reaction()`
        """
        warnings.warn("deprecated, alternative: bot.client.delete_reaction(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.delete_reaction(msg, emoji, user)

    async def list_game(self,
                        *,
                        begin_page: int = 1,
                        end_page: int = None,
                        page_size: int = 50,
                        sort: str = '') -> List[Game]:
        """list the games already registered at khl server

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.fetch_game_list()`"""
        warnings.warn("deprecated, alternative: bot.client.fetch_game_list(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.fetch_game_list(begin_page=begin_page,
                                                 end_page=end_page,
                                                 page_size=page_size,
                                                 sort=sort)

    async def create_game(self, name: str, process_name: str = None, icon: str = None) -> Game:
        """register a new game at khl server, can be used in profile status

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.register_game()`"""
        warnings.warn("deprecated, alternative: bot.client.register_game(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.register_game(name, process_name, icon)

    async def update_game(self, id: int, name: str = None, icon: str = None) -> Game:
        """update game already registered at khl server

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.update_game()`"""
        warnings.warn("deprecated, alternative: bot.client.update_game(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.client.update_game(id, name, icon)

    async def delete_game(self, game: Union[Game, int]):
        """unregister game from khl server

        :param game: accepts both Game object and bare game id(int type)

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.unregister_game()`"""
        warnings.warn("deprecated, alternative: bot.client.unregister_game(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.unregister_game(game)

    async def update_playing_game(self, game: Union[Game, int]):
        """update current playing game status

        :param game: accepts both Game object and bare id(int type)

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.update_playing_game()`"""
        warnings.warn("deprecated, alternative: bot.client.update_playing_game(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.update_playing_game(game)

    async def stop_playing_game(self):
        """clear current playing game status

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.stop_playing_game()`"""
        warnings.warn("deprecated, alternative: bot.client.stop_playing_game(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.stop_playing_game()

    async def update_listening_music(self, music_name: str, singer: str, software: Union[str, SoftwareTypes]):
        """update current listening music status

        :param music_name: name of music
        :param singer: singer of the music
        :param software: set software to playing the music

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.update_listening_music()`"""
        warnings.warn("deprecated, alternative: bot.client.update_listening_music(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.update_listening_music(music_name, singer, software)

    async def stop_listening_music(self):
        """clear current listening music status

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.stop_listening_music()`"""
        warnings.warn("deprecated, alternative: bot.client.stop_listening_music(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.stop_listening_music()

    async def update_channel(self,
                             channel: Union[str, PublicChannel],
                             name: str = None,
                             topic: str = None,
                             slow_mode: Union[int, SlowModeTypes] = None):
        """update channel's settings

        .. deprecated-removed:: 0.3.0 0.4.0
            use :func:`.client.update_channel()`"""
        warnings.warn("deprecated, alternative: bot.client.update_channel(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        await self.client.update_channel(channel, name, topic, slow_mode)

    async def start(self):
        for func in self._startup_index:
            await func(self)
        if self._is_running:
            raise RuntimeError('this bot is already running')
        self.task.schedule()
        await self.client.start()

    def run(self):
        """run the bot in blocking mode"""
        if not self.loop:
            self.loop = asyncio.get_event_loop()
        try:
            self.loop.run_until_complete(self.start())
        except KeyboardInterrupt:
            for func in self._shutdown_index:
                self.loop.run_until_complete(func(self))
            log.info('see you next time')



"""authorization/encrypt/decrypt works in khl"""
import base64
import json
from enum import Enum

from Cryptodome.Cipher import AES
from Cryptodome.Util import Padding


class Cert:
    """
    permission certification

    used in auth/data encrypt/decrypt
    """

    class Types(Enum):
        """
        types of :class:`Cert`

        used in extinguishing Cert and construct corresponding net client
        """
        NOTSET = 'not_set'
        WEBSOCKET = 'websocket'
        """
        websocket cert
        """
        WEBHOOK = 'webhook'
        """
        webhook cert
        """

    def __init__(self, *, type: Types = Types.NOTSET, token: str, verify_token: str = '', encrypt_key: str = ''):
        """
        all fields from bot config panel
        """
        if type != Cert.Types.NOTSET:
            self.type = type
        else:
            if verify_token:
                self.type = self.Types.WEBHOOK
            else:
                self.type = self.Types.WEBSOCKET
        self.token = token
        self.verify_token = verify_token
        self.encrypt_key = encrypt_key

    def decrypt(self, data: bytes) -> str:
        """ decrypt data

        :param data: encrypted byte array
        :return: decrypted str
        """
        if not self.encrypt_key:
            return ''
        data = base64.b64decode(data)
        data = AES.new(key=self.encrypt_key.encode().ljust(32, b'\x00'), mode=AES.MODE_CBC,
                       iv=data[0:16]).decrypt(base64.b64decode(data[16:]))
        data = Padding.unpad(data, 16)
        return data.decode('utf-8')

    def decode_raw(self, raw: bytes) -> dict:
        """decode raw package into plaintext data"""
        raw = json.loads(str(raw, encoding='utf-8'))
        return json.loads(self.decrypt(raw['encrypt'])) if ('encrypt' in raw.keys()) else raw



"""component context: carries environment context info for a message"""
from .channel import Channel
from .gateway import Requestable
from .guild import Guild


class Context(Requestable):
    """carries environment context info for a message"""
    channel: Channel
    guild: Guild

    def __init__(self, **kwargs):
        self.channel = kwargs.get('channel')
        self.guild = kwargs.get('guild')
        self.gate = kwargs.get('_gate_')



"""abstraction of khl client: handle to communicate with khl"""
import asyncio
import inspect
import logging
import time
from pathlib import Path
from typing import Dict, List, Callable, Coroutine, Union, IO

from . import api
from .channel import public_channel_factory, PublicChannel, Channel, PublicTextChannel
from .game import Game
from .gateway import Gateway, Requestable
from .guild import Guild, GuildBoost
from .interface import AsyncRunnable
from .message import RawMessage, Message, Event, PublicMessage, PrivateMessage
from ._types import SoftwareTypes, MessageTypes, SlowModeTypes
from .user import User
from .util import unpack_id, unpack_value

log = logging.getLogger(__name__)

TypeHandler = Callable[[Union['Message', 'Event']], Coroutine]


class Client(Requestable, AsyncRunnable):
    """
    The bridge between khl.py internal components and khl server.

    1. wraps net actions to khl backend, and translates raw objects into khl.py concepts/objects for internal use
    3. distributes package unwrapped by Receiver according to a handler map

    reminder: Client.loop only used to run handle_event() and registered handlers.
    """
    _handler_map: Dict[MessageTypes, List[TypeHandler]]

    def __init__(self, gate: Gateway):
        self.gate = gate
        self.ignore_self_msg = True
        self._me = None

        self._handler_map = {}
        self._pkg_queue = asyncio.Queue()

    def register(self, type: MessageTypes, handler: TypeHandler):
        """register handler to handle messages of type"""
        if not asyncio.iscoroutinefunction(handler):
            raise TypeError('handler must be a coroutine.')

        params = list(inspect.signature(handler).parameters.values())
        if len(params) != 1 or not issubclass(params[0].annotation, RawMessage):
            raise TypeError('handler must have one and only one param, and the param inherits RawMessage')

        if type not in self._handler_map:
            self._handler_map[type] = []
        self._handler_map[type].append(handler)

    async def handle_pkg(self):
        """consume `pkg` from `event_queue`"""
        while True:
            pkg: Dict = await self._pkg_queue.get()
            log.debug(f'upcoming pkg: {pkg}')

            try:
                await self._consume_pkg(pkg)
            except Exception as e:
                log.exception(e)

            self._pkg_queue.task_done()

    async def _consume_pkg(self, pkg: Dict):
        """
        spawn `msg` according to `pkg`,
        check if ignore msgs from self
        pass `msg` to corresponding handlers defined in `_handler_map`
        """
        msg = self._make_msg(pkg)
        if self.ignore_self_msg and msg.type != MessageTypes.SYS:
            if msg.author.id == (await self.fetch_me()).id:
                return
        self._dispatch_msg(msg)

    def _make_msg(self, pkg: Dict):
        if pkg.get('type') == MessageTypes.SYS.value:
            msg = Event(**pkg)
        else:
            msg = self._make_channel_msg(pkg)
        return msg

    def _make_channel_msg(self, pkg):
        msg = None
        channel_type = pkg.get('channel_type')
        if channel_type == 'GROUP':
            msg = PublicMessage(**pkg, _gate_=self.gate)
        elif channel_type == 'PERSON':
            msg = PrivateMessage(**pkg, _gate_=self.gate)
        else:
            log.error(f'can not make msg from pkg: {pkg}')
        return msg

    def _dispatch_msg(self, msg):
        if not msg:
            return
        handlers = self._handler_map.get(msg.type, ())
        for handler in handlers:
            asyncio.ensure_future(self._handle_safe(handler)(msg), loop=self.loop)

    @staticmethod
    def _handle_safe(handler: TypeHandler):

        async def safe_handler(msg):
            try:
                await handler(msg)
            except Exception as e:
                log.exception('error raised during message handling', exc_info=e)

        return safe_handler

    async def create_asset(self, file: Union[IO, str, Path]) -> str:
        """upload ``file`` to khl, and return the url to the file

        if ``file`` is a str or Path, ``open(file, 'rb')`` will be called to convert it into IO
        """
        if not isinstance(file, (str, Path)):
            return (await self.gate.exec_req(api.Asset.create(file=file)))['url']
        with open(file, 'rb') as f:
            return (await self.gate.exec_req(api.Asset.create(file=f)))['url']

    async def fetch_me(self, force_update: bool = False) -> User:
        """fetch detail of the ``User`` on the client"""
        if force_update or not self._me or not self._me.is_loaded():
            self._me = User(_gate_=self.gate, _lazy_loaded_=True, **(await self.gate.exec_req(api.User.me())))
        return self._me

    @property
    def me(self) -> User:
        """
        get client itself corresponding User

        RECOMMEND: use ``await fetch_me()``

        CAUTION: please call ``await fetch_me()`` first to load data from khl server

        designed as 'empty-then-fetch' will break the rule 'net-related is async'

        :return: the client's underlying User
        """
        if self._me and self._me.is_loaded():
            return self._me
        raise ValueError('not loaded, please call `await fetch_me()` first')

    async def fetch_public_channel(self, channel_id: str) -> PublicChannel:
        """fetch details of a public channel from khl"""
        channel_data = await self.gate.exec_req(api.Channel.view(channel_id))
        return public_channel_factory(_gate_=self.gate, **channel_data)

    async def fetch_user(self, user: Union[User, str]) -> User:
        """fetch detail of the specific user"""
        user_id = unpack_id(user)
        return User(_gate_=self.gate, _lazy_loaded_=True, **(await self.gate.exec_req(api.User.view(user_id))))

    async def fetch_guild(self, guild_id: str) -> Guild:
        """fetch details of a guild from khl"""
        guild = Guild(_gate_=self.gate, id=guild_id)
        await guild.load()
        return guild

    async def fetch_guild_list(self, **kwargs) -> List[Guild]:
        """list guilds which the client joined

        paged req, support standard pagination args"""
        guilds_data = (await self.gate.exec_paged_req(api.Guild.list(), **kwargs))
        return [Guild(_gate_=self.gate, _lazy_loaded_=True, **i) for i in guilds_data]

    async def leave(self, guild: Union[Guild, str]):
        """leave from ``guild``"""
        guild = Guild(_gate_=self.gate, id=guild) if isinstance(guild, str) else guild
        return await guild.leave()

    async def kickout(self, guild: Guild, user: Union[User, str]):
        """kick ``user`` out from ``guild``"""
        guild = Guild(_gate_=self.gate, id=guild) if isinstance(guild, str) else guild
        return await guild.kickout(user)

    async def delete_channel(self, channel: Union[Channel, str]):
        """delete a channel, permission required"""
        return await self.gate.exec_req(api.Channel.delete(unpack_id(channel)))

    @staticmethod
    async def send(target: Channel,
                   content: Union[str, List],
                   *,
                   type: MessageTypes = None,
                   temp_target_id: str = '',
                   **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP
        """
        if isinstance(target, PublicTextChannel):
            kwargs['temp_target_id'] = temp_target_id

        return await target.send(content, type=type, **kwargs)

    @staticmethod
    async def add_reaction(msg: Message, emoji: str):
        """add emoji to msg's reaction list

        wraps `Message.add_reaction`

        :param msg: accepts `Message`
        :param emoji: 😘
        """
        return await msg.add_reaction(emoji)

    @staticmethod
    async def delete_reaction(msg: Message, emoji: str, user: User = None):
        """delete emoji from msg's reaction list

        wraps `Message.delete_reaction`

        :param msg: accepts `Message`
        :param emoji: 😘
        :param user: whose reaction, delete others added reaction requires channel msg admin permission
        """
        return await msg.delete_reaction(emoji, user)

    async def fetch_game_list(self, **kwargs) -> List[Game]:
        """list the games already registered at khl server

        paged req, support standard pagination args"""
        games = await self.gate.exec_paged_req(api.game(), **kwargs)
        return [Game(**game_data) for game_data in games]

    async def register_game(self, name, process_name: str, icon: str) -> Game:
        """register a new game at khl server, can be used in profile status"""
        data = {
            'name': name,
        }
        if process_name is not None:
            data['process_name'] = process_name
        if icon is not None:
            data['icon'] = icon
        game_data = (await self.gate.exec_req(api.Game.create(**data)))
        return Game(**game_data)

    async def update_game(self, id: int, name: str, icon: str) -> Game:
        """update game already registered at khl server"""
        data = {'id': id}
        if name is not None:
            data['name'] = name
        if icon is not None:
            data['icon'] = icon
        game_data = (await self.gate.exec_req(api.Game.update(**data)))
        return Game(**game_data)

    async def unregister_game(self, game: Union[Game, int]):
        """unregister game from khl server

        :param game: accepts both Game object and bare game id(int type)
        """
        await self.gate.exec_req(api.Game.delete(id=unpack_id(game)))

    async def update_playing_game(self, game: Union[Game, int]):
        """update current playing game status

        :param game: accepts both Game object and bare id(int type)
        """
        await self.gate.exec_req(api.Game.activity(id=unpack_id(game), data_type=1))

    async def stop_playing_game(self):
        """clear current playing game status"""
        await self.gate.exec_req(api.Game.deleteActivity(data_type=1))

    async def update_listening_music(self, music_name: str, singer: str, software: Union[str, SoftwareTypes] = None):
        """update current listening music status

        :param music_name: such as 'Yesterday Once More'
        :param singer: such as 'Carpenters'
        :param software: music software, CLOUD_MUSIC as default
        """
        params = {'music_name': music_name, 'singer': singer, 'data_type': 2}
        if software:
            params['software'] = unpack_value(software)
        await self.gate.exec_req(api.Game.activity(**params))

    async def stop_listening_music(self):
        """clear current listening music status"""
        await self.gate.exec_req(api.Game.deleteActivity(data_type=2))

    async def update_channel(self,
                             channel: Union[str, PublicChannel],
                             name: str = None,
                             topic: str = None,
                             slow_mode: Union[int, SlowModeTypes] = None) -> PublicChannel:
        """update channel's settings"""
        channel = channel if isinstance(channel, PublicChannel) else await self.fetch_public_channel(channel)
        channel_data = await channel.update(name, topic, slow_mode)
        return public_channel_factory(_gate_=self.gate, **channel_data)

    async def fetch_guild_boost(self,
                                guild: Union[str, Guild],
                                start_time: int = 0,
                                end_time: int = int(time.time()),
                                **kwargs):
        """
        list the boost in guild.

        :param guild: guild_id or Guild object.
        :param start_time: start_time time stamp (Sec).
        :param end_time: end_time time stamp (Sec). Default to now time.
        """
        boost_list = await self.gate.exec_paged_req(
            api.GuildBoost.history(guild_id=unpack_id(guild), start_time=start_time, end_time=end_time),
            **kwargs)
        return [GuildBoost(**item, _gate_=self.gate) for item in boost_list]

    async def start(self):
        await asyncio.gather(self.handle_pkg(), self.gate.run(self._pkg_queue))



"""parser: component used in command args handling, convert string token to fit the command signature"""
import asyncio
import copy
import inspect
import logging
from typing import Dict, Any, Callable, List, Union

from khl import User, Channel, Client, Message, Role
from .exception import Exceptions
from .util import wrap_if_coro

log = logging.getLogger(__name__)


def _get_param(params: List[inspect.Parameter], index: int) -> Union[inspect.Parameter, None]:
    if len(params) <= 0:
        return None
    # for *args: result = the last param
    return params[min(index, len(params) - 1)]


def _get_param_type(param: Union[inspect.Parameter, None]):
    # str in default
    if param is None:
        return str
    if param.annotation == inspect.Parameter.empty:  # no type hint: str in default
        return str
    return param.annotation


async def _parse_user(_, client, token) -> User:
    if not (token.startswith("(met)") and token.endswith("(met)")):
        raise ValueError(f"wrong format: expected: '(met)`user_id`(met)', actual: {token}")
    return await client.fetch_user(token[5:-5])


async def _parse_channel(_, client, token) -> Channel:
    if not (token.startswith("(chn)") and token.endswith("(chn)")):
        raise ValueError(f"wrong format: expected: '(chn)`channel_id`(chn)', actual: {token}")
    return await client.fetch_public_channel(token[5:-5])


async def _parse_role(msg, _, token) -> Role:
    if not (token.startswith("(rol)") and token.endswith("(rol)")):
        raise ValueError(f"wrong format: expected: '(rol)`role_id`(rol)', actual: {token}")
    role_id = int(token[5:-5])
    role = next(filter(lambda r: r.id == role_id, await msg.ctx.guild.fetch_roles()), None)
    if role is None:
        raise ValueError("role not found in the guild")
    return role


class Parser:
    """
    deal with a list of tokens made from Lexer, convert their type to match the command.handler
    """
    _parse_funcs: Dict[Any, Callable] = {
        str: lambda msg, client, token: token,
        int: lambda msg, client, token: int(token),
        float: lambda msg, client, token: float(token),
        User: _parse_user,
        Channel: _parse_channel,
        Role: _parse_role
    }

    def __init__(self):
        self._parse_funcs = copy.copy(Parser._parse_funcs)

    async def parse(
        self,
        msg: Message,
        client: Client,
        tokens: List[str],
        params: List[inspect.Parameter],
    ) -> List[Any]:
        """
        parse tokens into args that types corresponding to handler's requirement

        :param msg: message object
        :param client: bot client
        :param tokens: output of Lexer.lex()
        :param params: command handlers parameters
        :return: List of args
        :raise: Parser.ArgListLenNotMatch
        """
        ret = []
        for i, token in enumerate(tokens):
            param = _get_param(params, i)
            param_type = _get_param_type(param)

            if param_type not in self._parse_funcs:
                raise Exceptions.Parser.NoParseFunc(param, token)
            func = self._parse_funcs[param_type]

            try:
                arg = await wrap_if_coro(func(msg, client, token))
            except Exception as e:
                raise Exceptions.Parser.ParseFailed(param, token, func, e)

            ret.append(arg)
        return ret

    def register(self, func):
        """
        decorator, register the func into object restricted _parse_funcs()

        checks if parse func for that type exists, and insert if not
        :param func: parse func
        """
        s = inspect.signature(func)

        # check: 1. not coroutine, 2. len matches
        if asyncio.iscoroutinefunction(func):
            raise TypeError('parse function should not be async')
        if len(s.parameters) != 1 or list(s.parameters.values())[0].annotation != str:
            raise TypeError('parse function should own only one param, and the param type is str')

        # insert, remember this is a replacement
        self._parse_funcs[s.return_annotation] = func
        return func



from typing import List

from . import api
from .user import User
from .interface import LazyLoadable
from .gateway import Requestable


class RolePermission:
    """part of channel permission, the permission setting for a specific role in the channel

    this setting overwrites the role's permission set in the guild"""
    role_id: int
    allow: int
    deny: int

    def __init__(self, *, role_id: int, allow: int, deny: int, **_):
        self.role_id = role_id
        self.allow = allow
        self.deny = deny


class UserPermission:
    """part of channel permission, the permission setting for a specific user"""
    user: User
    allow: int
    deny: int

    def __init__(self, *, user: User, allow: int, deny: int, **_):
        self.user = user
        self.allow = allow
        self.deny = deny


class ChannelPermission(LazyLoadable, Requestable):
    """permission settings in a channel, which can be customized.

    the custom permission entry can be divided into two types:
    1. customized role permission: overwrites the role's global permission settings
    2. customized user permission: exclusively set a user's permission in the channel
    """
    _id: str  # bound channel id
    _sync: int

    roles: List[RolePermission]
    users: List[UserPermission]

    @property
    def id(self) -> str:
        """which channel the permission belongs to"""
        return self._id

    @property
    def sync(self) -> bool:
        """if this channel's permission sync with category"""
        return self._sync != 0

    @sync.setter
    def sync(self, value: bool):
        self._sync = 1 if value else 0

    def __init__(self, **kwargs):
        self._id: str = kwargs.get('id')
        self.gate = kwargs.get('_gate_')
        self._load_fields(**kwargs)

    def _load_fields(self, **kwargs):
        self.roles = [RolePermission(**i) for i in kwargs.get('permission_overwrites', [])]
        self.users = [UserPermission(**i) for i in kwargs.get('permission_users', [])]
        self._sync = kwargs.get('permission_sync', None)
        if self.roles and self.users and (self._sync is not None):
            self._loaded = True

    async def load(self):
        self._load_fields(**await self.gate.exec_req(api.ChannelRole.index(channel_id=self.id)))



"""gateway related stuff"""
import asyncio
from abc import ABC
from typing import Union, List

from .api import _Req
from .receiver import Receiver
from .requester import HTTPRequester


class Gateway:
    """
    Component which deals with network connection and package send/receive

    reminder: this is not AsyncRunnable cuz gateway dose not have its own tasks, only pass loop to _in/_out
    """
    requester: HTTPRequester
    receiver: Receiver

    def __init__(self, requester: HTTPRequester, receiver: Receiver):
        self.requester = requester
        self.receiver = receiver

    async def request(self, method: str, route: str, **params) -> Union[dict, list]:
        """execute raw request, this is just a wrapper for convenience"""
        return await self.requester.request(method, route, **params)

    async def exec_req(self, r: _Req):
        """execute request, this is just a wrapper for convenience"""
        return await self.requester.exec_req(r)

    async def exec_paged_req(self, r: _Req, **kwargs) -> List:
        """execute paged request, this is just a wrapper for convenience"""
        return await self.requester.exec_paged_req(r, **kwargs)

    async def run(self, in_queue: asyncio.Queue):
        """run the receiver"""
        self.receiver.pkg_queue = in_queue
        await self.receiver.start()


class Requestable(ABC):
    """
    Classes that can use a `Gateway` to communicate with khl server.

    For example:
        `Message`: can use msg.reply() to send a reply to khl

        `Guild`: guild.get_roles() to fetch role list from khl
    """
    gate: Gateway



"""type enums"""
from enum import IntEnum, Enum
from typing import Dict, Any


class MessageTypes(IntEnum):
    """
    types of message, type==SYS will be interpreted as `Event`, others are `Message`
    """
    TEXT = 1
    IMG = 2
    VIDEO = 3
    FILE = 4
    AUDIO = 8
    KMD = 9
    CARD = 10
    SYS = 255


class ChannelTypes(IntEnum):
    """
    types of channel
    """
    CATEGORY = 0
    TEXT = 1
    VOICE = 2


class ChannelPrivacyTypes(Enum):
    """
    channel's privacy level
    """
    GROUP = 'GROUP'
    PERSON = 'PERSON'


class EventTypes(Enum):
    """
    for system events(type==255)
    """
    MESSAGE_BTN_CLICK = 'message_btn_click'

    ADDED_REACTION = 'added_reaction'
    DELETED_REACTION = 'deleted_reaction'
    UPDATED_MESSAGE = 'updated_message'
    DELETED_MESSAGE = 'deleted_message'
    MESSAGE_UPDATED = 'message_updated'

    PRIVATE_ADDED_REACTION = 'private_added_reaction'
    PRIVATE_DELETED_REACTION = 'private_deleted_reaction'
    UPDATED_PRIVATE_MESSAGE = 'updated_private_message'
    DELETED_PRIVATE_MESSAGE = 'deleted_private_message'

    UPDATED_GUILD = 'updated_guild'
    DELETED_GUILD = 'deleted_guild'
    ADDED_BLOCK_LIST = 'added_block_list'
    DELETED_BLOCK_LIST = 'deleted_block_list'

    ADDED_ROLE = 'added_role'
    DELETED_ROLE = 'deleted_role'
    UPDATED_ROLE = 'updated_role'

    JOINED_GUILD = 'joined_guild'
    EXITED_GUILD = 'exited_guild'
    GUILD_MEMBER_ONLINE = 'guild_member_online'
    GUILD_MEMBER_OFFLINE = 'guild_member_offline'

    UPDATED_GUILD_MEMBER = 'updated_guild_member'

    UPDATED_CHANNEL = 'updated_channel'
    ADDED_CHANNEL = 'added_channel'
    DELETED_CHANNEL = 'deleted_channel'

    JOINED_CHANNEL = 'joined_channel'
    EXITED_CHANNEL = 'exited_channel'
    USER_UPDATED = 'user_updated'
    SELF_JOINED_GUILD = 'self_joined_guild'
    SELF_EXITED_GUILD = 'self_exited_guild'

    PINNED_MESSAGE = 'pinned_message'
    UNPINNED_MESSAGE = 'unpinned_message'

    ADDED_EMOJI = 'added_emoji'
    DELETED_EMOJI = 'deleted_emoji'
    UPDATED_EMOJI = 'updated_emoji'


class GuildMuteTypes(IntEnum):
    """
    types of guild-mute
    """
    MIC = 1
    HEADSET = 2


class SlowModeTypes(IntEnum):
    """
    types of slow mode
    """
    FIVE_SEC = 5000
    TEN_SEC = 10000
    FIFTEEN_SEC = 15000
    THIRTY_SEC = 30000
    ONE_MIN = 60000
    TWO_MIN = 120000
    FIVE_MIN = 300000
    TEN_MIN = 600000
    FIFTEEN_MIN = 900000
    THIRTY_MIN = 1800000
    ONE_HOUR = 3600000
    TWO_HOUR = 7200000
    SIX_HOUR = 21600000

    @classmethod
    def possible_value(cls) -> Dict[Any, Enum]:
        """returns a dict with all enum names and corresponding values"""
        return cls._value2member_map_  # pylint: disable=no-member


class RoleTypes(IntEnum):
    """
    types of role type
    """
    UserCreated = 0
    BotSpecified = 1
    Booster = 2
    Everyone = 255


class SoftwareTypes(Enum):
    """types of listening music used software"""
    CLOUD_MUSIC = "cloudmusic"
    QQ_MUSIC = "qqmusic"
    KUGOU_MUSIC = "kugou"



"""khl raw API and related helpers"""
# pylint: skip-file
# this file contains hax to generate/interpolate stub code which make pylint pains
import functools
import inspect
import logging
import re
from collections import namedtuple
from typing import Callable, Tuple

import aiohttp

log = logging.getLogger(__name__)

_RE_ROUTE = re.compile(r'(?<!^)(?=[A-Z])')

_Req = namedtuple('_Req', ['method', 'route', 'params'])


def req(method: str, **http_fields):
    """meta-decorator

    :returns a decorator to fill func with boilerplate"""

    def _method(func: Callable):

        @functools.wraps(func)
        def req_maker(*args, **kwargs) -> _Req:
            route = _RE_ROUTE.sub('-', func.__qualname__).lower().replace('.', '/')

            # dump args into kwargs
            param_names = list(inspect.signature(func).parameters.keys())
            for i, arg in enumerate(args):
                kwargs[param_names[i]] = arg

            params = _merge_params(method, http_fields, kwargs)
            return _Req(method, route, params)

        return req_maker

    return _method


def _merge_params(method: str, http_fields: dict, req_args: dict) -> dict:
    payload = req_args
    payload_key = 'params'  # default payload_key: params=
    if method == 'POST':
        payload_key = 'json'  # POST: in default json=

        content_type = http_fields.get('headers', {}).get('Content-Type', None)
        if content_type == 'multipart/form-data':
            payload_key, payload = _build_form_payload(req_args)
            # headers of form-data req are delegated to aiohttp
            http_fields = _remove_content_type(http_fields)
        elif content_type is not None:
            raise ValueError(f'unrecognized Content-Type {content_type}')

    params = {payload_key: payload}
    params.update(http_fields)
    return params


def _remove_content_type(http_fields: dict) -> dict:
    """in some situation, such as content-type=multipart/form-data,
    content-type should be delegated to aiohttp to auto-generate,
    thus content-type is required to be removed in http_fields
    """
    if http_fields.get('headers', {}).get('Content-Type', None) is not None:
        http_fields = http_fields.copy()
        http_fields['headers'] = http_fields.get('headers', {}).copy()
        del http_fields['headers']['Content-Type']
    return http_fields


def _build_form_payload(req_args: dict) -> Tuple[str, aiohttp.FormData]:
    data = aiohttp.FormData()
    for name, value in req_args.items():
        data.add_field(name, value)
    return 'data', data


class Guild:
    """mapped to route: /guild/*"""

    @staticmethod
    @req('GET')
    def list():
        ...

    @staticmethod
    @req('GET')
    def view(guild_id):
        ...

    @staticmethod
    @req('GET')
    def userList(
        guild_id,
        channel_id=None,
        search=None,
        role_id=None,
        mobile_verified=None,
        active_time=None,
        joined_at=None,
        page=None,
        page_size=None,
    ):
        ...

    @staticmethod
    @req('POST')
    def nickname(
        guild_id,
        nickname,
        user_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def leave(guild_id):
        ...

    @staticmethod
    @req('POST')
    def kickout(guild_id, target_id):
        ...


class GuildMute:

    @staticmethod
    @req('GET')
    def list(guild_id, return_type):
        ...

    @staticmethod
    @req('POST')
    def create(
        guild_id,
        user_id,
        type,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
        guild_id,
        user_id,
        type,
    ):
        ...


class GuildBoost:

    @staticmethod
    @req('GET')
    def history(
            guild_id,
            start_time,
            end_time
    ):
        ...


class Blacklist:

    @staticmethod
    @req('GET')
    def list(
            guild_id
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
            guild_id,
            target_id,
            remark,
            del_msg_days
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
            guild_id,
            target_id
    ):
        ...


class Badge:

    @staticmethod
    @req('GET')
    def guild(
            guild_id,
            style
    ):
        ...


class Channel:

    @staticmethod
    @req('GET')
    def list(guild_id):
        ...

    @staticmethod
    @req('GET')
    def view(target_id):
        ...

    @staticmethod
    @req('POST')
    def create(guild_id, parent_id, name, type, limit_amount, voice_quality, is_category):
        ...

    @staticmethod
    @req('POST')
    def delete(channel_id):
        ...

    @staticmethod
    @req('POST')
    def moveUser(target_id, user_ids):
        ...

    @staticmethod
    @req('POST')
    def update(channel_id, name, topic, slow_mode):
        ...


class ChannelRole:

    @staticmethod
    @req('GET')
    def index(channel_id):
        ...

    @staticmethod
    @req('POST')
    def create(
        channel_id,
        type,
        value,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        channel_id,
        type,
        value,
        allow,
        deny,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
        channel_id,
        type,
        value,
    ):
        ...


class ChannelUser:

    @staticmethod
    @req('GET')
    def getJoinedChannel(page, page_size, guild_id, user_id):
        ...


class Message:

    @staticmethod
    @req('GET')
    def view(msg_id):
        ...

    @staticmethod
    @req('GET')
    def list(
        target_id,
        msg_id,
        pin,
        flag,
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
        type,
        target_id,
        content,
        quote,
        nonce,
        temp_target_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        msg_id,
        content,
        quote,
        temp_target_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(msg_id, ):
        ...

    @staticmethod
    @req('GET')
    def reactionList(msg_id, emoji):
        ...

    @staticmethod
    @req('POST')
    def addReaction(msg_id, emoji):
        ...

    @staticmethod
    @req('POST')
    def deleteReaction(msg_id, emoji, user_id):
        ...

    @staticmethod
    @req('POST')
    def addReaction(msg_id, emoji):
        ...

    @staticmethod
    @req('POST')
    def addReaction(msg_id, emoji):
        ...


class UserChat:

    @staticmethod
    @req('GET')
    def list():
        ...

    @staticmethod
    @req('GET')
    def view(chat_code):
        ...

    @staticmethod
    @req('POST')
    def create(target_id):
        ...

    @staticmethod
    @req('POST')
    def delete(chat_code):
        ...


class DirectMessage:

    @staticmethod
    @req('GET')
    def list(
        chat_code,
        target_id,
        msg_id,
        flag,
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
        type,
        target_id,
        chat_code,
        content,
        quote,
        nonce,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        msg_id,
        content,
        quote,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(msg_id):
        ...

    @staticmethod
    @req('GET')
    def reactionList(msg_id, emoji):
        ...

    @staticmethod
    @req('POST')
    def addReaction(msg_id, emoji):
        ...

    @staticmethod
    @req('POST')
    def deleteReaction(msg_id, emoji, user_id):
        ...


class Gateway:

    @staticmethod
    @req('GET')
    def index(compress):
        ...


class User:

    @staticmethod
    @req('GET')
    def me():
        ...

    @staticmethod
    @req('GET')
    def view(
        user_id,
        guild_id=None,
    ):
        ...

    @staticmethod
    @req('POST')
    def offline(
    ):
        ...


class Asset:

    @staticmethod
    @req('POST', headers={'Content-Type': 'multipart/form-data'})
    def create(file):
        ...


class GuildRole:

    @staticmethod
    @req('GET')
    def list(guild_id):
        ...

    @staticmethod
    @req('POST')
    def create(
        name,
        guild_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        guild_id,
        role_id,
        hoist,
        mentionable,
        permissions,
        color,
        name,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
        guild_id,
        role_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def grant(
        guild_id,
        user_id,
        role_id,
    ):
        ...

    @staticmethod
    @req('POST')
    def revoke(
        guild_id,
        user_id,
        role_id,
    ):
        ...


class Intimacy:

    @staticmethod
    @req('GET')
    def index(user_id):
        ...

    @staticmethod
    @req('POST')
    def update(
        user_id,
        score,
        social_info,
        img_id,
    ):
        ...


class GuildEmoji:

    @staticmethod
    @req('GET')
    def list(guild_id):
        ...

    @staticmethod
    @req('POST', headers={'Content-Type': 'multipart/form-data'})
    def create(
        name,
        guild_id,
        emoji,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        name,
        id,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(id):
        ...


class Invite:

    @staticmethod
    @req('GET')
    def list(
        guild_id,
        channel_id,
        page,
        page_size
    ):
        ...

    @staticmethod
    @req('POST')
    def create(
        guild_id,
        channel_id,
        duration,
        setting_times
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(
        url_code,
        guild_id,
        channel_id
    ):
        ...


@req('GET')
def game():
    ...


class Game:

    @staticmethod
    @req('POST')
    def create(
        name,
        process_name,
        icon,
    ):
        ...

    @staticmethod
    @req('POST')
    def update(
        id,
        name,
        icon,
    ):
        ...

    @staticmethod
    @req('POST')
    def delete(id):
        ...

    @staticmethod
    @req('POST')
    def activity(data_type, id=None, music_name=None, singer=None, software=None):
        ...

    @staticmethod
    @req('POST')
    def deleteActivity(data_type):
        ...


class Oauth2:

    @staticmethod
    @req("POST")
    def token(
        grant_type,
        client_id,
        client_secret,
        code,
        redirect_uri
    ):
        ...



from typing import Union, Callable, Coroutine, Any

from khl import Message, User

TypeRule = Callable[[Message], Union[bool, Coroutine[Any, Any, bool]]]


class Rule:
    """check if the msg fulfills some condition"""

    @staticmethod
    def is_bot_mentioned(bot) -> TypeRule:
        """:return: a Rule that checks if the bot is mentioned"""

        async def rule(msg: Message) -> bool:
            return (await bot.fetch_me()).id in msg.extra.get('mention')

        return rule

    @staticmethod
    def is_user_mentioned(user: User) -> TypeRule:
        """:return: a Rule that checks if the user is mentioned"""

        def rule(msg: Message) -> bool:
            return user.id in msg.extra.get('mention')

        return rule

    @staticmethod
    def is_mention_all(msg: Message) -> bool:
        """:return: a Rule that checks if the msg mentioned all members"""
        return msg.extra.get('mention_all', None) is not None

    @staticmethod
    def is_not_bot(msg: Message) -> bool:
        """:return: a Rule that check if the msg belong non bot"""
        return not msg.author.bot



import json
from typing import List, Union

from . import api
from .gateway import Requestable, Gateway
from .interface import LazyLoadable
from ._types import MessageTypes


class User(LazyLoadable, Requestable):
    """
    `Standard Object`

    represent a entity that interact with khl server
    """
    id: str
    username: str
    nickname: str
    identify_num: str
    online: bool
    bot: bool
    status: int
    avatar: str
    vip_avatar: str
    mobile_verified: bool

    _loaded: bool
    gate: Gateway

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', '')
        self.username = kwargs.get('username', '')
        self.nickname = kwargs.get('nickname', '')
        self.identify_num = kwargs.get('identify_num', '')
        self.online = kwargs.get('online', False)
        self.bot = kwargs.get('bot', False)
        self.status = kwargs.get('status', 0)
        self.avatar = kwargs.get('avatar', '')
        self.vip_avatar = kwargs.get('vip_avatar', '')
        self.mobile_verified = kwargs.get('mobile_verified', False)

        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_', None)

    async def load(self):
        pass

    async def send(self, content: Union[str, List], *, type: MessageTypes = None, **kwargs):
        """
        send a msg to a channel

        ``temp_target_id`` is only available in ChannelPrivacyTypes.GROUP
        """
        # if content is card msg, then convert it to plain str
        if isinstance(content, List):
            type = MessageTypes.CARD
            content = json.dumps(content)
        else:
            type = type or MessageTypes.KMD

        # merge params
        kwargs['target_id'] = self.id
        kwargs['content'] = content
        kwargs['type'] = type.value

        return await self.gate.exec_req(api.DirectMessage.create(**kwargs))



"""khl.py main import lists"""

# infra
from .interface import AsyncRunnable, LazyLoadable
from ._types import (MessageTypes, ChannelTypes, ChannelPrivacyTypes, EventTypes, GuildMuteTypes, SlowModeTypes,
                     RoleTypes, SoftwareTypes)
from .cert import Cert
from .receiver import Receiver, WebhookReceiver, WebsocketReceiver
from .requester import HTTPRequester
from .gateway import Gateway, Requestable
from .client import Client

# concepts
from .role import Role
from .user import User
from .channel import Channel, PublicTextChannel, PrivateChannel, PublicChannel
from .guild import Guild, GuildUser
from .context import Context
from .message import RawMessage, Message, PublicMessage, PrivateMessage, Event

# extensions
from .bot import Bot



"""all Type Enums and related helpers in card message"""
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union, Dict, List


class _Representable(ABC):
    """internal interface, can be cast to JSON"""

    @property
    @abstractmethod
    def _repr(self) -> Union[str, Dict, List]:
        """cast class object to JSON serializable representation"""
        raise NotImplementedError


class _TypeEnum(Enum):
    """base class of all types(involved in card components)

    remind: TypeEnum implements _repr but not inherits from Representable,
    since "TypeError: metaclass conflict:
    the metaclass of a derived class must be a (non-strict) subclass of the
    metaclasses of all its bases"
    """

    @property
    def _repr(self):
        return self.value


class Types:
    """contains all types used in card messages"""

    class Theme(_TypeEnum):
        """describes a component's theme, controls its color"""
        NA = ''
        PRIMARY = 'primary'
        SECONDARY = 'secondary'
        SUCCESS = 'success'
        DANGER = 'danger'
        WARNING = 'warning'
        INFO = 'info'
        NONE = 'none'

    class Size(_TypeEnum):
        """describes a component's size"""
        NA = ''
        XS = 'xs'
        SM = 'sm'
        MD = 'md'
        LG = 'lg'

    class Text(_TypeEnum):
        """which type of text"""
        PLAIN = 'plain-text'
        KMD = 'kmarkdown'

    class Click(_TypeEnum):
        """used in button, determines the behavior of button when clicked"""
        LINK = 'link'
        RETURN_VAL = 'return-val'

    class SectionMode(_TypeEnum):
        """used in section, arrangement of elements in"""
        LEFT = 'left'
        RIGHT = 'right'

    class File(_TypeEnum):
        """which type of file"""
        FILE = 'file'
        AUDIO = 'audio'
        VIDEO = 'video'

    class CountdownMode(_TypeEnum):
        """used in countdown module, determines its layout"""
        DAY = 'day'
        HOUR = 'hour'
        SECOND = 'second'


def _get_repr(item) -> Union[str, Dict, List]:
    """a helper function for serialization"""
    if isinstance(item, list):
        return [_get_repr(i) for i in item]
    return getattr(item, '_repr', item)


class _Common(_Representable, ABC):
    _type: str
    _theme: Types.Theme
    _size: Types.Size

    def __init__(self, theme: Union[Types.Theme, str, None], size: Union[Types.Size, str, None]):
        super().__init__()
        self._theme = Types.Theme(theme) if theme else None
        self._size = Types.Size(size) if size else None

    def _gen_dict(self, field_list: List) -> Dict:
        result = {}
        for k in field_list:
            # get repr of k/_k(private field with exported key)
            obj = _get_repr(getattr(self, k, None)) or _get_repr(getattr(self, '_' + k, None))
            if obj is not None:
                result[k] = obj
        return result


class _Module(_Common, ABC):
    ...


class _Struct(_Common, ABC):
    ...



"""common hax"""


def unpack_id(obj):
    """extract obj's id if not basic data type"""
    return obj if isinstance(obj, (str, int)) else obj.id


def unpack_value(enum):
    """extract enum's value if not basic data type"""
    return enum if isinstance(enum, (str, int)) else enum.value



import logging

from apscheduler.events import EVENT_JOB_ERROR
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

from khl import AsyncRunnable

log = logging.getLogger(__name__)


class TaskManager(AsyncRunnable):
    """manage tasks"""
    _scheduler: AsyncIOScheduler

    def __init__(self):
        self._scheduler = AsyncIOScheduler()

    def add_interval(self,
                     weeks=0,
                     days=0,
                     hours=0,
                     minutes=0,
                     seconds=0,
                     start_date=None,
                     end_date=None,
                     timezone=None,
                     jitter=None):
        """decorator, add a interval type task"""
        trigger = IntervalTrigger(weeks=weeks,
                                  days=days,
                                  hours=hours,
                                  minutes=minutes,
                                  seconds=seconds,
                                  start_date=start_date,
                                  end_date=end_date,
                                  timezone=timezone,
                                  jitter=jitter)
        return lambda func: self._scheduler.add_job(func, trigger)

    def add_cron(self,
                 year=None,
                 month=None,
                 day=None,
                 week=None,
                 day_of_week=None,
                 hour=None,
                 minute=None,
                 second=None,
                 start_date=None,
                 end_date=None,
                 timezone=None,
                 jitter=None):
        """decorator, add a cron type task"""
        trigger = CronTrigger(year=year,
                              month=month,
                              day=day,
                              week=week,
                              day_of_week=day_of_week,
                              hour=hour,
                              minute=minute,
                              second=second,
                              start_date=start_date,
                              end_date=end_date,
                              timezone=timezone,
                              jitter=jitter)
        return lambda func: self._scheduler.add_job(func, trigger)

    def add_date(self, run_date=None, timezone=None):
        """decorator, add a date type task"""
        trigger = DateTrigger(run_date=run_date, timezone=timezone)
        return lambda func: self._scheduler.add_job(func, trigger)

    async def start(self):
        self._scheduler.configure({'event_loop': self.loop}, '')
        self._scheduler.add_listener(lambda e: log.exception('error raised during task', exc_info=e.exception),
                                     EVENT_JOB_ERROR)
        self._scheduler.start()  # reminder: this is not blocking

    @property
    def scheduler(self):
        """getter, get the scheduler"""
        return self._scheduler



import asyncio
import logging
from typing import Union, List

from aiohttp import ClientSession

from .api import _Req
from .cert import Cert

log = logging.getLogger(__name__)

API = 'https://www.kookapp.cn/api/v3'


class HTTPRequester:
    """wrap raw requests, handle boilerplate param filling works"""

    def __init__(self, cert: Cert):
        self._cert = cert
        self._cs: ClientSession = ClientSession()

    def __del__(self):
        asyncio.get_event_loop().run_until_complete(self._cs.close())

    async def request(self, method: str, route: str, **params) -> Union[dict, list, bytes]:
        """wrap raw request, fill authorization, handle & extract response"""
        headers = params.pop('headers', {})
        params['headers'] = headers

        log.debug(f'{method} {route}: req: {params}')  # token is excluded
        headers['Authorization'] = f'Bot {self._cert.token}'
        async with self._cs.request(method, f'{API}/{route}', **params) as res:
            if res.content_type == 'application/json':
                rsp = await res.json()
                if rsp['code'] != 0:
                    raise HTTPRequester.APIRequestFailed(method, route, params, rsp['code'], rsp['message'])
                rsp = rsp['data']
            else:
                rsp = await res.read()
            log.debug(f'{method} {route}: rsp: {rsp}')
            return rsp

    async def exec_req(self, r: _Req):
        """_Req -> raw request"""
        return await self.request(r.method, r.route, **r.params)

    async def exec_paged_req(self,
                             r: _Req,
                             *,
                             begin_page: int = 1,
                             end_page: int = None,
                             page_size: int = 50,
                             sort: str = '') -> List:
        """
        execute paged requests

        iter from ``begin_page`` to the ``end_page``, ``end_page=None`` means to the end

        1. get a req, inject the params
        2. req and receive the result
        3. unwrap the result, append result, fresh pagination params

        :param begin_page: int = 1,
        :param end_page: int = None,
        :param page_size: int = 50,
        :param sort: str = ''
        """
        ret = []
        current_page = begin_page
        while end_page is None or current_page <= end_page:
            r.params['params']['page'] = current_page
            r.params['params']['page_size'] = page_size
            if sort:
                r.params['params']['sort'] = sort

            p = await self.exec_req(r)

            ret.extend(p['items'])
            current_page = p['meta']['page']
            page_total = p['meta']['page_total']
            page_size = p['meta']['page_size']

            current_page += 1
            if end_page is None:
                end_page = page_total

        return ret

    class APIRequestFailed(Exception):
        """Raised when khl.py received non-zero error code from remote server.

        By default, params (request body) is not included when calling __str__ on it, to avoid leaking credential
        into logs;

        if request body is needed for debug purpose, consider explicitly catching this exception and
        call repr(...) with the exception instance."""

        def __init__(self, method, route, params, err_code, err_message):
            super().__init__()
            self.method = method
            self.route = route
            self.params = params
            self.err_code = err_code
            self.err_message = err_message

        def __str__(self):
            return f"Requesting '{self.method} {self.route}' failed with {self.err_code}: {self.err_message}"



import logging
import re
import shlex
from abc import ABC, abstractmethod
from typing import List, Set, Union, Pattern

from khl import Message
from .exception import Exceptions

log = logging.getLogger(__name__)


class Lexer(ABC):
    """
    deal with msg, convert msg.content -> a list of tokens(like tokens in compiler)
    """

    @abstractmethod
    def lex(self, msg: Message) -> List[str]:
        """
        read msg, and get a list of tokens as List[str] for following parser.parse()

        :param msg: which will be torn down into List[str]
        :return: Tuple(is the msg match the lexer, thrived tokens)
        """
        raise NotImplementedError


class DefaultLexer(Lexer):
    """
    lex with python standard module ``shlex``
    """
    prefixes: Set[str]
    triggers: Set[str]

    case_sensitive: bool

    def __init__(self, prefixes: Set[str], triggers: Set[str], case_sensitive: bool):
        self.prefixes = prefixes
        self.triggers = triggers
        self.case_sensitive = case_sensitive
        if not self.case_sensitive:
            self.triggers = set(map(lambda t: t.lower(), self.triggers))

    def lex(self, msg: Message) -> List[str]:
        """
        checks if msg starts with any prefix, shlex.split() if hit, then check if starts with any trigger

        :param msg: which will be torn down into List[str]
        :return: Tuple(is the msg match the lexer, thrived tokens)
        :raises:    ShlexLexer.MalformedContent if shlex.split() failed
                    ShlexLexer.NoMatchedPrefix  literally
                    ShlexLexer.NoMatchedTrigger literally
        """
        matched_prefixes = [p for p in self.prefixes if msg.content.startswith(p)]
        if not matched_prefixes:
            raise Exceptions.Lexer.NotMatched()

        for prefix in matched_prefixes:
            try:
                arg_list = shlex.split(msg.content[len(prefix):])
            except Exception as e:
                raise DefaultLexer.MalformedContent(msg) from e
            a0 = arg_list[0] if len(arg_list) > 0 else ''
            if (a0 if self.case_sensitive else a0.lower()) not in self.triggers:
                raise Exceptions.Lexer.NotMatched()
            return arg_list[1:]  # arg_list[0] is trigger

    class MalformedContent(Exceptions.Lexer.LexFailed):
        """the content can not be shlex.split()"""


class RELexer(Lexer):
    """
    lex with python built-in module ``re``
    """
    pattern: Pattern

    def __init__(self, regex: Union[str, Pattern]):
        if isinstance(regex, str):
            self.pattern = re.compile(regex)
        else:
            self.pattern = regex

    def lex(self, msg: Message) -> List[str]:
        m = self.pattern.fullmatch(msg.content)
        if not m:
            raise Exceptions.Lexer.NotMatched()
        return [m[i] for i in range(1, len(m.groups()) + 1) if m.start(i) < len(msg.content)]



"""guild related stuffs: Guild, ChannelCategory"""
import logging
import time
import warnings
from typing import List, Union, Dict, IO

from . import api
from .channel import Channel, public_channel_factory, PublicChannel, PublicVoiceChannel
from .gateway import Requestable, Gateway
from .interface import LazyLoadable
from .role import Role
from ._types import ChannelTypes, GuildMuteTypes
from .user import User
from .util import unpack_id

log = logging.getLogger(__name__)


class GuildUser(User):
    """a user in guild

    with some fields more than User"""
    guild_id: str
    joined_at: int
    active_time: int
    roles: List[int]
    gate: Gateway

    def __init__(self, **kwargs):
        self.roles = kwargs.get('roles', [])
        self.guild_id = kwargs.get('guild_id', '')
        self.joined_at = kwargs.get('joined_at', 0)
        self.active_time = kwargs.get('active_time', 0)
        super().__init__(**kwargs)

    async def fetch_roles(self, **kwargs) -> List[Role]:
        """
        Get the user roles in this guild

        paged req, support standard pagination args

        :return: A list for Role
        """
        guild_roles = (await self.gate.exec_paged_req(api.GuildRole.list(self.guild_id), **kwargs))
        rt: List[Role] = []
        for role in guild_roles:
            if role['role_id'] in self.roles:
                rt.append(Role(**role))
        return rt


class GuildBoost:
    """Guild boost"""
    user_id: str
    guild_id: str
    start_time: int
    end_time: int
    user: User

    def __init__(self, **kwargs) -> None:
        self.user_id = kwargs.get('user_id')
        self.guild_id = kwargs.get('guild_id')
        self.start_time = kwargs.get('start_time')
        self.end_time = kwargs.get('end_time')
        self.user = User(**kwargs.get('user'), _gate_=kwargs.get('_gate_', None))


class ChannelCategory(Requestable):
    """represent a channel set"""
    id: str
    name: str
    master_id: str
    guild_id: str
    level: int
    limit_amount: int
    _channels: List[PublicChannel]

    def __init__(self, **kwargs):
        self.gate = kwargs.get('_gate_', None)
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.master_id = kwargs.get('master_id')
        self.guild_id = kwargs.get('_guild_id_')
        self.level = kwargs.get('level')
        self.limit_amount = kwargs.get('limit_amount')
        self._channels = kwargs.get('channels', [])

    def append(self, *channel: PublicChannel):
        """append var-len channel(s) into this category"""
        self._channels.append(*channel)

    def pop(self, index: int = None) -> PublicChannel:
        """pop a channel(default last) from this category"""
        return self._channels.pop(index)

    async def create_channel(self,
                             name: str,
                             type: ChannelTypes = None,
                             limit_amount: int = None,
                             voice_quality: int = None) -> PublicChannel:
        """create a channel in this channel category

        docs: https://developer.kaiheila.cn/doc/http/channel#%E5%88%9B%E5%BB%BA%E9%A2%91%E9%81%93"""
        params = {'name': name, 'guild_id': self.guild_id, 'parent_id': self.id}
        if type is not None:
            params['type'] = type.value
        if limit_amount:
            params['limit_amount'] = limit_amount
        if voice_quality:
            params['voice_quality'] = voice_quality
        pc = public_channel_factory(self.gate, **(await self.gate.exec_req(api.Channel.create(**params))))
        self._channels.append(pc)
        return pc

    async def delete_channel(self, channel: Union[Channel, str]):
        """delete a channel from this channel category"""
        channel_id = channel.id if isinstance(channel, Channel) else channel
        if channel_id not in [i.id for i in self._channels]:
            raise ValueError(f'channel {channel_id} is not belongs to this category')
        return await self.gate.exec_req(api.Channel.delete(channel_id))

    def __iter__(self):
        return iter(self._channels)


class Guild(LazyLoadable, Requestable):
    """
    `Standard Object`

    represent a server where users gathered in and contains channels
    """
    id: str
    name: str
    topic: str
    master_id: str
    icon: str
    notify_type: int
    region: str
    enable_open: bool
    open_id: str
    default_channel_id: str
    welcome_channel_id: str
    _roles: List[Role]
    _channel_categories: Dict[str, ChannelCategory]
    _channels: List[PublicChannel]

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self._channel_categories = {}
        self._loaded = kwargs.get('_lazy_loaded_', False)
        self.gate = kwargs.get('_gate_', None)
        self._update_fields(**kwargs)

    def _update_fields(self, **kwargs):
        self.name = kwargs.get('name', '')
        self.topic = kwargs.get('topic', '')
        self.master_id = kwargs.get('master_id', '')
        self.icon = kwargs.get('icon', '')
        self.notify_type = kwargs.get('notify_type', 0)
        self.region = kwargs.get('region', '')
        self.enable_open = kwargs.get('enable_open', False)
        self.open_id = kwargs.get('open_id', '')
        self.default_channel_id = kwargs.get('default_channel_id', '')
        self.welcome_channel_id = kwargs.get('welcome_channel_id', '')
        self._roles = kwargs.get('roles', None)
        self._channels = kwargs.get('channels', None)

    async def load(self):
        self._update_fields(**(await self.gate.exec_req(api.Guild.view(self.id))))
        self._loaded = True

    async def fetch_channel_category_list(self, force_update: bool = True) -> List[ChannelCategory]:
        """fetch all channel category as a list"""
        await self.fetch_channel_list(force_update)
        return list(self._channel_categories.values())

    async def fetch_channel_list(self, force_update: bool = True) -> List[PublicChannel]:
        """fetch channel list from khl server"""
        if force_update or self._channels is None:
            raw_list = await self.gate.exec_paged_req(api.Channel.list(guild_id=self.id))
            channels: List[PublicChannel] = []
            channel_categories: Dict[str, ChannelCategory] = {}
            for i in raw_list:
                if i['is_category']:
                    cc = ChannelCategory(_gate_=self.gate, _guild_id_=self.id, **i)
                    channel_categories[cc.id] = cc
                else:
                    channels.append(public_channel_factory(_gate_=self.gate, **i))

            self._channels = []
            for i in channels:
                if i.parent_id in channel_categories:
                    channel_categories[i.parent_id].append(i)
                else:
                    self._channels.append(i)
            self._channel_categories = channel_categories
        return self._merge_channels()

    def _merge_channels(self) -> List[PublicChannel]:
        channels = []
        channels.extend(self._channels)
        for v in self._channel_categories.values():
            channels.extend(v)
        return channels

    @property
    def channels(self) -> List[Channel]:
        """
        get guild's channel list

        RECOMMEND: use ``await fetch_channel_list()``

        CAUTION: please call ``await fetch_me()`` first to load data from khl server

        designed as 'empty-then-fetch' will break the rule 'net-related is async'
        """
        if self.loaded:
            return self._merge_channels()
        raise ValueError('not loaded, please call `await fetch_channel_list()` first')

    async def list_user(self, channel: Union[Channel, str] = None, **kwargs) -> List[User]:
        """list users in the guild/a channel belongs to the guild

        paged req, support standard pagination args

        .. deprecated-removed:: 0.3.2 0.4.0
            use :func:`fetch_user_list()`"""
        warnings.warn("deprecated, alternative: fetch_user_list(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return await self.fetch_user_list(channel, **kwargs)

    async def fetch_user_list(self, channel: Union[Channel, str] = None, **kwargs) -> List[User]:
        """list users in the guild/a channel belongs to the guild

        paged req, support standard pagination args"""
        cid = channel.id if isinstance(channel, Channel) else channel
        params = {'guild_id': self.id}
        if cid is not None:
            params['channel_id'] = cid
        users = await self.gate.exec_paged_req(api.Guild.userList(**params), **kwargs)
        return [User(_gate_=self.gate, _lazy_loaded_=True, **i) for i in users]

    async def fetch_joined_channel(self,
                                   user: Union[User, str],
                                   page: int = 1,
                                   page_size: int = 50) -> List[PublicVoiceChannel]:
        """fetch the channels which the user joined(public voice channel)"""
        channels = await self.gate.exec_paged_req(
            api.ChannelUser.getJoinedChannel(page=page, page_size=page_size, guild_id=self.id, user_id=unpack_id(user)))
        return [PublicVoiceChannel(_gate_=self.gate, _lazy_loaded_=True, **i) for i in channels]

    async def fetch_user(self, user_id: str) -> GuildUser:
        """get user object from user_id, can only fetch user in current guild
        """
        user = await self.gate.exec_req(api.User.view(user_id=user_id, guild_id=self.id))
        return GuildUser(guild_id=self.id, _gate_=self.gate, _lazy_loaded_=True, **user)

    async def set_user_nickname(self, user: Union[User, str], new_nickname: str):
        """set the user's nickname in this guild"""
        await self.gate.exec_req(api.Guild.nickname(guild_id=self.id, nickname=new_nickname, user_id=unpack_id(user)))

    async def fetch_roles(self, force_update: bool = True) -> List[Role]:
        """fetch the role list in the guild"""
        if force_update or self._roles is None:
            raw_list = await self.gate.exec_paged_req(api.GuildRole.list(guild_id=self.id))
            self._roles = [Role(**i) for i in raw_list]
        return self._roles

    async def create_role(self, role_name: str) -> Role:
        """create a role in the guild"""
        return Role(**(await self.gate.exec_req(api.GuildRole.create(guild_id=self.id, name=role_name))))

    async def update_role(self, new_role: Role) -> Role:
        """update a role in the guild

        :param new_role an edited role object"""
        return Role(**(await self.gate.exec_req(api.GuildRole.update(guild_id=self.id, **vars(new_role)))))

    async def delete_role(self, role: Union[int, Role]):
        """delete a role from the guild"""
        return await self.gate.exec_req(api.GuildRole.delete(guild_id=self.id, role_id=unpack_id(role)))

    async def grant_role(self, user: Union[User, str], role: Union[Role, int]):
        """
        docs:
        https://developer.kaiheila.cn/doc/http/guild-role#%E8%B5%8B%E4%BA%88%E7%94%A8%E6%88%B7%E8%A7%92%E8%89%B2
        """
        return await self.gate.exec_req(
            api.GuildRole.grant(guild_id=self.id, user_id=unpack_id(user), role_id=unpack_id(role)))

    async def revoke_role(self, user: Union[User, str], role: Union[Role, int]):
        """
        docs:
        https://developer.kaiheila.cn/doc/http/guild-role#%E5%88%A0%E9%99%A4%E7%94%A8%E6%88%B7%E8%A7%92%E8%89%B2
        """
        return await self.gate.exec_req(
            api.GuildRole.revoke(guild_id=self.id, user_id=unpack_id(user), role_id=unpack_id(role)))

    async def create_channel(self,
                             name: str,
                             type: ChannelTypes = None,
                             category: Union[str, ChannelCategory] = None,
                             limit_amount: int = None,
                             voice_quality: int = None):
        """create a channel in the guild

        docs: https://developer.kaiheila.cn/doc/http/channel#%E5%88%9B%E5%BB%BA%E9%A2%91%E9%81%93"""
        params = {'name': name, 'guild_id': self.id}
        if type is not None:
            if type == ChannelTypes.CATEGORY:
                params['is_category'] = 1
            else:
                params['type'] = type.value
        if category:
            params['parent_id'] = unpack_id(category)
        if limit_amount:
            params['limit_amount'] = limit_amount
        if voice_quality:
            params['voice_quality'] = voice_quality
        return public_channel_factory(self.gate, **(await self.gate.exec_req(api.Channel.create(**params))))

    async def delete_channel(self, channel: Union[Channel, str]):
        """delete the channel from the guild"""
        return await self.gate.exec_req(api.Channel.delete(unpack_id(channel)))

    async def kickout(self, user: Union[User, str]):
        """kick the user from the guild"""
        return await self.gate.exec_req(api.Guild.kickout(guild_id=self.id, target_id=unpack_id(user)))

    async def leave(self):
        """leave from this guild"""
        return await self.gate.exec_req(api.Guild.leave(guild_id=self.id))

    async def get_mute_list(self, return_type: str = 'detail'):
        """get mute list from this guild

        .. deprecated-removed:: 0.3.2 0.4.0
            use :func:`fetch_mute_list()`"""
        warnings.warn("deprecated, alternative: fetch_mute_list(), everything else is in the same",
                      DeprecationWarning,
                      stacklevel=2)
        return self.fetch_mute_list(return_type)

    async def fetch_mute_list(self, return_type: str = 'detail'):
        """get mute list from this guild"""
        return await self.gate.exec_req(api.GuildMute.list(guild_id=self.id, return_type=return_type))

    async def create_mute(self, user: Union[User, str], type: GuildMuteTypes):
        """create mute on this guild"""
        user_id = user.id if isinstance(user, User) else user
        return await self.gate.exec_req(api.GuildMute.create(guild_id=self.id, user_id=user_id, type=type.value))

    async def delete_mute(self, user: Union[User, str], type: GuildMuteTypes):
        """delete mute from this guild"""
        user_id = user.id if isinstance(user, User) else user
        return await self.gate.exec_req(api.GuildMute.delete(guild_id=self.id, user_id=user_id, type=type.value))

    async def fetch_emoji_list(self) -> List[Dict]:
        """fetch guild emoji list

        :returns: a list of emoji dict, dict contains {'name', 'id', 'user_info': who uploaded the emoji}
        """
        return await self.gate.exec_paged_req(api.GuildEmoji.list(guild_id=self.id))

    async def create_emoji(self, emoji: Union[IO, str], *, name: str = None) -> Dict:
        """upload a custom emoji to the guild

        :returns: an emoji dict. For emoji dict structure, please refer to fetch_emoji_list() doc
        """
        if isinstance(emoji, str):
            emoji = open(emoji, 'rb')
        params = {'guild_id': self.id, 'emoji': emoji}
        if name is not None:
            params['name'] = name
        return await self.gate.exec_req(api.GuildEmoji.create(**params))

    async def update_emoji(self, id: str, *, name: str = None):
        """update a custom emoji's name"""
        params = {'id': id}
        if name is not None:
            params['name'] = name
        return await self.gate.exec_req(api.GuildEmoji.update(**params))

    async def delete_emoji(self, id: str):
        """delete a custom emoji"""
        return await self.gate.exec_req(api.GuildEmoji.delete(id))

    async def fetch_boost(self, start_time: int = 0, end_time: int = int(time.time()), **kwargs) -> List[GuildBoost]:
        """
        list the boost in guild.

        :param start_time: start_time time stamp (Sec).
        :param end_time: end_time time stamp (Sec). Default to now time.
        """
        boost_list = await self.gate.exec_paged_req(
            api.GuildBoost.history(guild_id=self.id, start_time=start_time, end_time=end_time), **kwargs)
        return [GuildBoost(**item, _gate_=self.gate) for item in boost_list]
